// d-lib_preset.cpp
// Preset and support functions
#pragma once

/////////////
// DEFINES //
/////////////

#define PRE_PARAMS 256
#define ED_SCRNS 20
#define ED_SCRN_PARAMS 8
#define ED_SCRN_COLS 4
#define ED_SCRN_ROWS 5
#define LCD_COLS 2
#define LCD_ROWS 4


///////////////
// CONSTANTS //
///////////////

const string DLP_EXT = "dlp";
const string BNK_EXT = "bnk";
const string SPI_EXT = "spi";
const string EEPROM_EXT = "eeprom";


struct param_tlp {
	uint32_t type;  // type
	string lbl;  // label
	string ptr;  // pointer name
};

const param_tlp SYS_TLP[] = {
	{   1, "50Hz",  "s_p0_ds"},	// 0
	{   7, "Dith",  "s_p1_ds"},	// 1
	{   3, "P<>V",  "s_p2_ds"},	// 2
	{   1, "Erev",  "s_p3_ds"},	// 3
	{   7, "Dith",  "s_p4_ds"},	// 4
	{0x24, "LCD ",  "s_p5_ds"},	// 5
	{0xcb, "Pcal",  "p_p0_ds"},	// 6
	{0xc0, "Lin ",  "p_p1_ds"},	// 7
	{0x26, "Ofs-",  "p_p2_ds"},	// 8
	{0xcb, "Sens",  "p_p3_ds"},	// 9
	{0x26, "Ofs+",  "p_p4_ds"},	// 10
	{0xfe, "Cent",  "p_p5_ds"},	// 11
	{0xcb, "Vcal",  "v_p0_ds"},	// 12
	{0xc0, "Lin ",  "v_p1_ds"},	// 13
	{0x26, "Ofs-",  "v_p2_ds"},	// 14
	{0xcb, "Sens",  "v_p3_ds"},	// 15
	{0x34, "Drop",  "v_p4_ds"},	// 16
	{0x26, "Ofs+",  "v_p5_ds"},	// 17
	{0x35, "Vol ",  "v_p6_ds"},	// 18
	{   1, "Out ",  "v_p7_ds"},	// 19
	{0x20, "LED ",  "t_p0_ds"},	// 20
	{   4, "Qant",  "t_p1_ds"},	// 21
	{   3, "Post",  "t_p2_ds"},	// 22
	{0x9b, "Note",  "t_p3_ds"},	// 23
	{0x9f, "Oct ",  "t_p4_ds"},	// 24
	{0xc5, "Bass", "eq_p0_ds"},	// 25
	{0xc5, "Treb", "eq_p1_ds"},	// 26
	{0x35, "Line",  "v_p8_ds"},	// 27
	{0x7e, "Wait",  "s_p6_ds"},	// 28
	{   1, "Trak",  "p_p6_ds"},	// 29
	{0x7e, "Auto",  "p_p7_ds"},	// 30
	{   0, "Load", "ps_p0_ds"},	// 31 - NOT sys / stored / recalled!
	{   0, "Stor", "ps_p1_ds"},	// 32 - NOT sys / stored / recalled!
	{   0, "Load", "ps_p2_ds"},	// 33 - NOT sys / stored / recalled!
	{   0, "Stor", "ps_p3_ds"}	// 34 - NOT sys / stored / recalled!
};

const param_tlp USR_TLP[] = {
	{0x35, "osc ",  "o_p0_ds"},	// 0
	{0x24, "odd ",  "o_p1_ds"},	// 1
	{0x24, "harm",  "o_p2_ds"},	// 2
	{0xca, "pmod",  "o_p3_ds"},	// 3
	{0xca, "vmod",  "o_p4_ds"},	// 4
	{0x97, "oct ",  "o_p5_ds"},	// 5
	{0xf0, "offs",  "o_p6_ds"},	// 6
	{0x24, "xmix",  "o_p7_ds"},	// 7
	{0x24, "fm  ",  "o_p8_ds"},	// 8
	{0x70, "freq",  "o_p9_ds"},	// 9
	{0x76, "reso", "o_p10_ds"},	// 10
	{0x94, "mode", "o_p11_ds"},	// 11
	{0xca, "pmod", "o_p12_ds"},	// 12
	{0xca, "vmod", "o_p13_ds"},	// 13
	{0xc5, "bass", "o_p14_ds"},	// 14
	{0xc5, "treb", "o_p15_ds"},	// 15
	{0xf0, "hmul", "o_p16_ds"},	// 16
	{0xf0, "hmul", "o_p17_ds"},	// 17
	{0xf0, "offs", "o_p18_ds"},	// 18
	{0x35, "sprd", "o_p19_ds"},	// 19
	{0x24, "xmix", "o_p20_ds"},	// 20
	{0x74, "nois",  "n_p0_ds"},	// 21
	{0x70, "freq",  "n_p3_ds"},	// 22
	{0x76, "reso",  "n_p4_ds"},	// 23
	{0x94, "mode",  "n_p5_ds"},	// 24
	{0xca, "pmod",  "n_p6_ds"},	// 25
	{0xca, "vmod",  "n_p7_ds"},	// 26
	{0xca, "pmod",  "n_p8_ds"},	// 27
	{0xca, "vmod",  "n_p9_ds"},	// 28
	{0x34, "puls", "n_p10_ds"},	// 29
	{0xc5, "bass", "n_p11_ds"},	// 30
	{0x24, "xmix", "n_p12_ds"},	// 31
	{0xc5, "treb", "n_p13_ds"},	// 32
	{0x24, "duty", "n_p14_ds"},	// 33
	{0xf1, "reso",  "r_p0_ds"},	// 34
	{0xc6, "harm",  "r_p1_ds"},	// 35
	{0x72, "freq",  "r_p2_ds"},	// 36
	{0xc6, "tap ",  "r_p3_ds"},	// 37
	{0x71, "hpf ",  "r_p4_ds"},	// 38
	{0xc5, "xmix",  "r_p5_ds"},	// 39
	{0x92, "mode",  "r_p6_ds"},	// 40
	{0x70, "freq",  "f_p0_ds"},	// 41
	{0x35, "levl",  "f_p1_ds"},	// 42
	{0x70, "freq",  "f_p2_ds"},	// 43
	{0x35, "levl",  "f_p3_ds"},	// 44
	{0x70, "freq",  "f_p4_ds"},	// 45
	{0x35, "levl",  "f_p5_ds"},	// 46
	{0xca, "pmod",  "f_p6_ds"},	// 47
	{0xca, "vmod",  "f_p7_ds"},	// 48
	{0xca, "pmod",  "f_p8_ds"},	// 49
	{0xca, "vmod",  "f_p9_ds"},	// 50
	{0xca, "pmod", "f_p10_ds"},	// 51
	{0xca, "vmod", "f_p11_ds"},	// 52
	{0x76, "reso", "f_p12_ds"},	// 53
	{0x76, "reso", "f_p13_ds"},	// 54
	{0x70, "freq", "f_p14_ds"},	// 55
	{0x35, "levl", "f_p15_ds"},	// 56
	{0x70, "freq", "f_p16_ds"},	// 57
	{0x35, "levl", "f_p17_ds"},	// 58
	{0x70, "freq", "f_p18_ds"},	// 59
	{0x35, "levl", "f_p19_ds"},	// 60
	{0x76, "reso", "f_p20_ds"},	// 61
	{0x70, "freq", "f_p22_ds"},	// 62
	{0x35, "levl", "f_p23_ds"},	// 63
	{0x70, "freq", "f_p24_ds"},	// 64
	{0x35, "levl", "f_p25_ds"},	// 65
	{0xca, "pmod", "f_p26_ds"},	// 66
	{0xca, "vmod", "f_p27_ds"},	// 67
	{0x76, "reso", "f_p28_ds"},	// 68
	{0x24, "cntr", "pc_p0_ds"},	// 69
	{0x24, "slew", "pc_p1_ds"},	// 70
	{0x44, "span", "pc_p2_ds"},	// 71
	{0x24, "levl", "pc_p3_ds"},	// 72
	{0xc6, "vloc", "pc_p4_ds"},	// 73
	{0x25, "kloc",  "e_p0_ds"},	// 74
	{0x44, "knee",  "e_p1_ds"},	// 75
	{0x76, "fall",  "e_p2_ds"},	// 76
	{0x75, "rise",  "e_p3_ds"},	// 77
	{0x34, "velo",  "e_p4_ds"},	// 78
	{0x73, "damp",  "e_p5_ds"},	// 79
	{0x25, "dloc",  "e_p6_ds"},	// 80
	{0x74, "prev", "pp_p0_ds"},	// 81
	{0xc5, "harm", "pp_p1_ds"},	// 82
	{0x97, "oct ", "pp_p2_ds"},	// 83
	{0xca, "pmod", "pp_p3_ds"},	// 84
	{0xca, "vmod", "pp_p4_ds"},	// 85
	{0xc5, "treb", "pp_p5_ds"},	// 86
	{0xc5, "bass", "pp_p6_ds"},	// 87
	{0xb0, "chan",  "m_p0_ds"},	// 88
	{0x25, "vloc",  "m_p1_ds"},	// 89
	{0x42, "bend",  "m_p2_ds"},	// 90
	{0x97, "oct ",  "m_p3_ds"},	// 91
	{0x30, "velo",  "m_p4_ds"},	// 92
	{0xf2, "cc  ",  "m_p5_ds"},	// 93
	{0x45, "cloc",  "m_p6_ds"},	// 94
	{0x8b, "prvw", "pp_p7_ds"}	// 95
};
	
const string ED_PTRS[ED_SCRNS][ED_SCRN_PARAMS] = {
	{"v_p6_ds",   "v_p7_ds",  "v_p0_ds",   "p_p0_ds", "ps_p1_ds",   "o_p5_ds",  "ps_p0_ds", "    D-LEV"},
	{"v_p6_ds",   "v_p8_ds", "pp_p0_ds",  "eq_p1_ds",  "o_p0_ds",  "eq_p0_ds",   "n_p0_ds", "   LEVELS"},
	{"pp_p4_ds", "pp_p3_ds", "pp_p0_ds",  "pp_p5_ds", "pp_p1_ds",  "pp_p6_ds",  "pp_p2_ds", "  PREVIEW"},
	{"m_p1_ds",   "m_p4_ds",  "m_p6_ds",   "m_p5_ds",  "m_p2_ds",   "m_p0_ds",   "m_p3_ds", "     MIDI"},
	{"e_p0_ds",   "e_p3_ds",  "e_p1_ds",   "e_p2_ds",  "e_p4_ds",   "e_p5_ds",   "e_p6_ds", "   VOLUME"},
	{"pc_p4_ds", "pc_p3_ds", "pc_p1_ds",  "pc_p2_ds", "pc_p0_ds",  "pp_p7_ds",   "t_p2_ds", "    PITCH"},
	{"n_p9_ds",   "n_p8_ds",  "n_p0_ds",  "n_p13_ds", "n_p10_ds",  "n_p11_ds",  "n_p14_ds", "    NOISE"},
	{"n_p7_ds",   "n_p6_ds",  "n_p3_ds",   "n_p0_ds",  "n_p5_ds",  "n_p12_ds",   "n_p4_ds", "FLT_NOISE"},
	{"o_p4_ds",   "o_p3_ds",  "o_p2_ds",  "o_p15_ds",  "o_p1_ds",  "o_p14_ds",   "o_p5_ds", "    0_OSC"},
	{"o_p6_ds",  "o_p18_ds", "o_p16_ds",  "o_p17_ds",  "o_p8_ds",  "o_p19_ds",   "o_p7_ds", "    1_OSC"},
	{"o_p13_ds", "o_p12_ds",  "o_p9_ds",   "o_p0_ds", "o_p11_ds",  "o_p20_ds",  "o_p10_ds", "  FLT_OSC"},
	{"r_p2_ds",   "r_p3_ds",  "r_p4_ds",   "r_p1_ds",  "r_p6_ds",   "r_p5_ds",   "r_p0_ds", "    RESON"},
	{"f_p7_ds",   "f_p6_ds",  "f_p0_ds",   "f_p1_ds", "f_p14_ds",  "f_p15_ds",  "f_p12_ds", "   0_FORM"},
	{"f_p9_ds",   "f_p8_ds",  "f_p2_ds",   "f_p3_ds", "f_p16_ds",  "f_p17_ds",  "f_p13_ds", "   1_FORM"},
	{"f_p11_ds", "f_p10_ds",  "f_p4_ds",   "f_p5_ds", "f_p22_ds",  "f_p23_ds",  "f_p20_ds", "   2_FORM"},
	{"f_p27_ds", "f_p26_ds", "f_p18_ds",  "f_p19_ds", "f_p24_ds",  "f_p25_ds",  "f_p28_ds", "   3_FORM"},
	{"v_p0_ds",   "v_p4_ds",  "v_p1_ds",   "s_p4_ds",  "v_p2_ds",   "v_p5_ds",   "v_p3_ds", "  V_FIELD"},
	{"p_p0_ds",   "p_p6_ds",  "p_p1_ds",   "s_p1_ds",  "p_p2_ds",   "p_p4_ds",   "p_p3_ds", "  P_FIELD"},
	{"t_p0_ds",   "p_p5_ds",  "s_p5_ds",   "t_p3_ds",  "t_p1_ds",   "t_p4_ds",   "t_p2_ds", "  DISPLAY"},
	{"s_p6_ds",   "p_p7_ds",  "s_p2_ds",   "s_p0_ds", "ps_p3_ds",   "s_p3_ds",  "ps_p2_ds", "   SYSTEM"}
};

// this is bs!
const int32_t SYS_PARAMS = sizeof(SYS_TLP)/sizeof(*SYS_TLP);
const int32_t USR_PARAMS = sizeof(USR_TLP)/sizeof(*USR_TLP);

// editor preset type
struct ed_pre_t {
	int32_t sys_enc[PRE_PARAMS] = {};
	int32_t usr_enc[PRE_PARAMS] = {};
};

// zero out preset
void ed_pre_cla(ed_pre_t& ed_pre_i) {
	fill(begin(ed_pre_i.sys_enc), end(ed_pre_i.sys_enc), 0);
	fill(begin(ed_pre_i.usr_enc), end(ed_pre_i.usr_enc), 0);
	return;
}

// return encoder maximum value
int32_t enc_max(uint32_t type_i) {
	int32_t max = 0;  // default
	if (type_i < 0x20) { max = type_i & 0x1f; }  // 0 thru 31
	else if (type_i < 0x70) { max = (1 << ((type_i & 0x3) + 5)) - 1; }  // 31, 63, 127, 255
	else if (type_i < 0x72) { max = 192; }  // 192
	else if (type_i == 0x72) { max = 127; }  // 127
	else if (type_i < 0x78) { max = 63; }  // 63
	else if (type_i == 0x7e) { max = 99; }  // 99
	else if (type_i == 0x7f) { max = 19; }  // 19
	else if (type_i < 0xc0) { max = type_i & 0x1f; }  // 0 thru 31
	else if (type_i < 0xf0) { max = (1 << ((type_i & 0x3) + 4)) - 1; }  // 15, 31, 63, 127
	else if (type_i == 0xf0) { max = 127; }  // 127
	else if (type_i == 0xf1) { max = 63; }  // 63
	else if (type_i == 0xf2) { max = 31; }  // 31
	else if (type_i == 0xfe) { max = 99; }  // 99
	else { max = 0; }
	return(max);
}

// return encoder minimum value
int32_t enc_min(uint32_t type_i) {
	int32_t min = 0;  // default
	if (type_i < 0x80) { min = 0; }  // 0
	else if (type_i < 0xa0) { min = -(enc_max(type_i) + 1); }  // -1, -2, ..., -16
	else if (type_i == 0xf2) { min = -127; }  // -127
	else { min = -enc_max(type_i); }
	return(min);
}

// return encoder signed status
bool enc_signed(uint32_t type_i) {
	if (enc_min(type_i) < 0) { return(true); }
	else { return(false); }
}

// return true if idx_i out of range
bool idx_error(int32_t idx_i, bool usr_i) {
	int32_t idx_max = (usr_i) ? USR_PARAMS : SYS_PARAMS;
	if ((idx_i < 0) || (idx_i > idx_max)) { return(true); }
	else { return(false); }
}

// return sys / usr label value
// return empty string if idx_i out of range
string lbl_idx_rd(int32_t idx_i, bool usr_i) {
	if (idx_error(idx_i, usr_i)) { return(""); }
	if (usr_i) { return(USR_TLP[idx_i].lbl); }
	else { return(SYS_TLP[idx_i].lbl); }
}

// return sys / usr type value
// return 0 if idx_i out of range
uint32_t type_idx_rd(int32_t idx_i, bool usr_i) {
	if (idx_error(idx_i, usr_i)) { return(0); }
	if (usr_i) { return(USR_TLP[idx_i].type); }
	else { return(SYS_TLP[idx_i].type); }
}

// return sys / usr encoder value
// return 0 if idx_i out of range
int32_t enc_idx_rd(const ed_pre_t& ed_pre_i, int32_t idx_i, bool usr_i) {
	if (idx_error(idx_i, usr_i)) { return(0); }
	if (usr_i) { return(ed_pre_i.usr_enc[idx_i]); }
	else { return(ed_pre_i.sys_enc[idx_i]); }
}

// limit encoder to min / max
// return input if idx_i out of range
int32_t enc_lim(int32_t enc_i, int32_t idx_i, bool usr_i) {
	if (idx_error(idx_i, usr_i)) { return(enc_i); }
	uint32_t type = type_idx_rd(idx_i, usr_i);
	int32_t min = enc_min(type);
	int32_t max = enc_max(type);
	if (enc_i > max) { return(max); }
	else if (enc_i < min) { return(min); }
	else { return(enc_i); }
}

// write sys / usr encoder value to editor preset
// return true if error
bool enc_idx_wr(ed_pre_t& ed_pre_io, int32_t enc_i, int32_t idx_i, bool usr_i) {
	if (idx_error(idx_i, usr_i)) { return(true); }
	int32_t enc_tmp = enc_lim(enc_i, idx_i, usr_i);
	if (usr_i) { ed_pre_io.usr_enc[idx_i] = enc_tmp; }
	else { ed_pre_io.sys_enc[idx_i] = enc_tmp; }
	return(false);
}

// return encoder index and group given ptr string
// return -1 if no match
int32_t ptr_to_idx(const string& ptr_i, bool& usr_o) {
	bool match_f = false;  // flag
	// search sys params
	usr_o = false;  // default (sys)
	int32_t idx = 0;  // init
	while (!match_f && (idx < SYS_PARAMS)) {
		if (ptr_i == SYS_TLP[idx].ptr) { match_f = true; }
		else { idx++; }
	}
	if (match_f) { return(idx); }
	// else, search user params
	usr_o = true;
	idx = 0;
	while (!match_f && (idx < USR_PARAMS)) {
		if (ptr_i == USR_TLP[idx].ptr) { match_f = true; }
		else { idx++; }
	}
	if (match_f) { return(idx); }
	else { return(-1); }
}

// return filter freq value (type 0xf0, 0xf1)
// 7041 * EXP2((ENC * (2^27) / 24) + 3/4)
// input: [0:192]
// output: 27 to 7040 (Hz)
int32_t filt_freq(int32_t enc_i) {
	uint32_t mult = pow(2, 27) / 24;
	uint32_t offs = 0xc0000000;
	uint32_t enc_mo = (enc_i * mult) + offs;
	return(7041 * (pow(2, (enc_mo / pow(2, 27))) / pow(2, 32)));
}

// return nearest encoder value for a given filter frequency
uint8_t filt_enc(uint64_t freq_i) {
	uint32_t error_best = 1<<31;  // some big value
	uint32_t enc_o = 0;
	for(int i=0; i<=192; i++) {
		uint32_t freq = filt_freq(i);  // test freq
		uint32_t error = labs(int64_t(freq_i) - int64_t(freq));
		if (error < error_best) {
			error_best = error;
			enc_o = i;
		}
	}
	return(enc_o);
}

// return reson freq value (type 0xf2)
// 48001 / ((((((~(ENC<<25))^4)*0.871)+((~(ENC<<25))>>3))>>22)+4)
// input: [0:127]
// output: 46 to 9600 (Hz)
int32_t reson_freq(int32_t enc_i) {
	uint32_t enc_fs_rev = ~(enc_i << 25);
	uint64_t efr = enc_fs_rev;
	uint64_t sq = (efr * efr) >> 32;
	uint64_t qd = (sq * sq) >> 32;
	return(48001 / ((((uint32_t(qd * 0.871) + (efr / 8)) >> 22) + 4)));
}

// return nearest encoder value for a given reson frequency
uint8_t reson_enc(uint64_t freq_i) {
	uint32_t error_best = 1<<31;  // some big value
	uint32_t enc_o = 0;
	for(int i=0; i<=127; i++) {
		uint32_t freq = reson_freq(i);  // test freq
		uint32_t error = labs(int64_t(freq_i) - int64_t(freq));
		if (error < error_best) {
			error_best = error;
			enc_o = i;
		}
	}
	return(enc_o);
}

// return sys / usr encoder display string
string enc_disp_idx_rd(const ed_pre_t& ed_pre_i, int32_t idx_i, bool usr_i) {
	stringstream str_o;
	int32_t rtn = 0;
	int32_t enc = enc_idx_rd(ed_pre_i, idx_i, usr_i);
	int32_t type = type_idx_rd(idx_i, usr_i);
	if ((type == 0x70) || (type == 0x71)) { rtn = filt_freq(enc); }
	else if (type == 0x72) { rtn = reson_freq(enc); }
	else { rtn = enc; }
	if (type == 0) { str_o << setw(5) << "";}
	else { str_o << setw(5) << rtn; }
	return(str_o.str());
}

// return true if x_i out of range
bool ed_x_error(int32_t x_i) { return((x_i < 0) || (x_i >= (ED_SCRN_COLS * LCD_COLS))); }

// return true if y_i out of range
bool ed_y_error(int32_t y_i) { return((y_i < 0) || (y_i >= (ED_SCRN_ROWS * LCD_ROWS))); }

// return mod x
int32_t mod_ed_x(int32_t x_i) { return(mod_int(x_i, ED_SCRN_COLS * LCD_COLS)); }
	
// return mod y
int32_t mod_ed_y(int32_t y_i) { return(mod_int(y_i, ED_SCRN_ROWS * LCD_ROWS)); }

// return scrn given UI(x_i,y_i)
int32_t ed_xy_to_scrn(int32_t x_i, int32_t y_i) {
	return((x_i / LCD_COLS) + ((y_i / LCD_ROWS) * LCD_ROWS));
}

// return param given UI(x_i,y_i)
int32_t ed_xy_to_param(int32_t x_i, int32_t y_i) {
	return((x_i % LCD_COLS) + ((y_i % LCD_ROWS) * LCD_COLS));
}

// return true if UI(x_i,y_i) at LCD screen title
bool ed_xy_on_title(int32_t x_i, int32_t y_i) {
	if ((mod_int(x_i, LCD_COLS) == LCD_COLS - 1) && (mod_int(y_i, LCD_ROWS) == LCD_ROWS - 1)) { return(true); }
	else { return(false); }
}

// move x/y position
void ed_xy_move(int32_t& x_io, int32_t& y_io, int32_t dx_i, int32_t dy_i) {
	int32_t x_max = (ED_SCRN_COLS * LCD_COLS) - 1;
	int32_t y_max = (ED_SCRN_ROWS * LCD_ROWS) - 1;
	// offset
	x_io += dx_i;
	y_io += dy_i;
	// limit
	if (x_io < 0) { x_io = 0; }
	if (x_io > x_max) { x_io = x_max; }
	if (y_io < 0) { y_io = 0; }
	if (y_io > y_max) { y_io = y_max; }
	return;
}

// move x/y position to home
void ed_xy_home(int32_t& x_io, int32_t& y_io) {
	x_io = LCD_COLS - 1;
	y_io = LCD_ROWS - 1;
	return;
}

// move x/y position to end
void ed_xy_end(int32_t& x_io, int32_t& y_io) {
	x_io = (ED_SCRN_COLS * LCD_COLS) - 1;
	y_io = (ED_SCRN_ROWS * LCD_ROWS) - 1;
	return;
}

// draw editor UI screen
void ed_draw(const ed_pre_t& ed_pre_io, int32_t ed_x_i, int32_t ed_y_i, bool ed_i) {
	int32_t idx = 0;
	bool usr_f = true;
	for (int32_t y=0; y<ED_SCRN_ROWS*LCD_ROWS; y++) {
		if (!(y % LCD_ROWS)) { con_font('h'); cout << setw(CON_W) << ""; con_font('n'); cout << endl; }  // row divider
		for (int32_t x=0; x<ED_SCRN_COLS*LCD_COLS; x++) {
			int32_t scrn = ed_xy_to_scrn(x, y);
			int32_t param = ed_xy_to_param(x, y);
			if (!(x % LCD_COLS)) { con_font('h'); cout << " "; con_font('n'); }  // col divider
			if (param == 7) {  // draw screen title
				if ((x == ed_x_i) && (y == ed_y_i) && ed_i) { con_font('s'); }  // hilite
				else { con_font('o'); }
				cout << ED_PTRS[scrn][param]; 
				con_font('n'); 
			}
			else {  // draw screen parameter
				idx = ptr_to_idx(ED_PTRS[scrn][param], usr_f);
				if ((x == ed_x_i) && (y == ed_y_i) && ed_i) { con_font('s'); }  // hilite
				cout << lbl_idx_rd(idx, usr_f);
				cout << enc_disp_idx_rd(ed_pre_io, idx, usr_f);
				con_font('n');
				if (!(x % LCD_COLS)) { cout << "  "; }
			}
		}
		con_font('h'); cout << " " << endl; con_font('n');  // ending col divider
	}
	con_font('o'); cout << setw(50) << right << "F1 - EDITOR" << setw(35) << ""; 
	con_font('n'); 
	cout << endl;  // ending row divider
	return;
}

// print editor UI screen => txt string
string ed_print_to_str(const ed_pre_t& ed_pre_io) {
	string h_line_sub = '+' + string(22, '-');
	string h_line = " " + h_line_sub + h_line_sub + h_line_sub + h_line_sub + "+ \n";
	stringstream ss;
	int32_t idx = 0;
	bool usr_f = true;
	for (int32_t y=0; y<ED_SCRN_ROWS*LCD_ROWS; y++) {
		if (!(y % LCD_ROWS)) { ss << h_line; }  // row divider
		for (int32_t x=0; x<ED_SCRN_COLS*LCD_COLS; x++) {
			int32_t scrn = ed_xy_to_scrn(x, y);
			int32_t param = ed_xy_to_param(x, y);
			if (!(x % LCD_COLS)) { ss << " | "; }  // col divider
			if (param == 7) { ss << ED_PTRS[scrn][param]; }
			else { 
				idx = ptr_to_idx(ED_PTRS[scrn][param], usr_f);
				ss << lbl_idx_rd(idx, usr_f);
				ss << enc_disp_idx_rd(ed_pre_io, idx, usr_f);
				con_font('n');
				if (!(x % LCD_COLS)) { ss << "  "; }
			}
		}
		ss << " | " << endl;  // ending col divider
	}
	ss << h_line;  // ending row divider
	return(ss.str());
}

// print editor UI screen => txt file
// return true if error
bool ed_print_to_file(const ed_pre_t& ed_pre_io, const string& tfname_i, const string& psname_i, bool append_i) {
	string str = ed_print_to_str(ed_pre_io);
	bool error_f = str_to_file(tfname_i, str, append_i);
	if (!error_f) {
		stringstream ss;
		ss << "preset: " << psname_i << endl << endl;
		error_f = str_to_file(tfname_i, ss.str(), true);
	}
	return(error_f);
}

// preset string => preset unsigned byte array
// return true if error
bool pre_str_to_pre_uba(uint8_t pre_uba_o[], const string& pre_str_i) {
	bool error_f = false;
	uint32_t idx = 0;
	stringstream ss(pre_str_i);
	string line = "";
	while (!error_f && getline(ss, line)) {
		uint64_t line_hex = 0;  // default
		error_f = !str_is_hex(line);
		if (!error_f) { 
			line_hex = strtoul(line.c_str(), nullptr, 16);  // STRTOUL!
			for (uint32_t b=0; b<4; b++) {
				uint64_t line_byte = line_hex & 0xff;
				pre_uba_o[idx] = uint8_t(line_byte);
				line_hex >>= 8;
				idx++;
			}
		}
		if (idx > PRE_PARAMS) { error_f = true; }  // too many params
	}
	if (idx != PRE_PARAMS) { error_f = true; }  // s/b 256 params
	return(error_f);
}


// preset unsigned byte array => editor
void pre_uba_to_ed(ed_pre_t& ed_pre_io, uint8_t pre_uba_i[], bool usr_i) {
	uint32_t idx = 0;
	while (idx < PRE_PARAMS) {
		int32_t enc = 0;
		if (enc_signed(type_idx_rd(idx, usr_i))) { enc = int8_t(pre_uba_i[idx]); }
		else { enc = uint8_t(pre_uba_i[idx]); }
		if (usr_i) { ed_pre_io.usr_enc[idx] = enc; }
		else { ed_pre_io.sys_enc[idx] = enc; }
		idx++;
	}
	return;
}

// preset file => editor
// return true if error
bool dlp_to_ed(ed_pre_t& ed_pre_io, const string& dlp_i, bool usr_i) {
	string pre_str = "";
	uint8_t pre_uba[PRE_PARAMS] = { 0 };
	bool error_f = file_to_str(dlp_i, pre_str);
	if (!error_f) { error_f = pre_str_to_pre_uba(pre_uba, pre_str); }
	if (!error_f) { pre_uba_to_ed(ed_pre_io, pre_uba, usr_i); }
	return(error_f);
}

// preset unsigned byte array => preset string
// return true if error
void pre_uba_to_pre_str(uint8_t pre_uba_i[], string& pre_str_o) {
	uint32_t idx = 0;
	stringstream ss_o;
	while (idx < PRE_PARAMS) {
		uint32_t line_hex = 0;
		for (uint32_t b=0; b<4; b++) {
			uint8_t enc = pre_uba_i[idx];
			line_hex += enc << (b * 8);
			idx++;
		}
		ss_o << hex << line_hex << endl;
	}
	pre_str_o = ss_o.str();
	return;
}

void ed_to_pre_uba(const ed_pre_t& ed_pre_io, uint8_t pre_uba_o[], bool usr_i) {
	uint32_t idx = 0;
	while (idx < PRE_PARAMS) {
		if (usr_i) { pre_uba_o[idx] = ed_pre_io.usr_enc[idx]; }
		else { pre_uba_o[idx] = ed_pre_io.sys_enc[idx]; }
		idx++;
	}
	return;
}

// preset string => preset file
// return true if error
bool pre_str_to_dlp(const string& pre_str_i, const string& dlp_i ) {
	ofstream wr_file;
	wr_file.open(dlp_i);
	bool error_f = !wr_file.is_open();
	if (!error_f) {
		stringstream ss_o(pre_str_i);
		wr_file << ss_o.rdbuf();
	}
	if (wr_file.is_open()) { wr_file.close(); }  // close file if open
	return(error_f);
}

// editor => preset file
// return true if error
bool ed_to_dlp(const ed_pre_t& ed_pre_io, const string& dlp_i, bool usr_i) {
	uint8_t pre_uba[PRE_PARAMS] = { 0 };
	string pre_str = "";
	ed_to_pre_uba(ed_pre_io, pre_uba, usr_i);
	pre_uba_to_pre_str(pre_uba, pre_str);
	return(pre_str_to_dlp(pre_str, dlp_i));
}

// memory byte write
// return true if error
bool mem_byte_wr(const sp_type& sp_i, uint32_t addr_i, int32_t byte_i) {
	bool error_f = sp_open_err(sp_i);  // check port
	if (!error_f) {
		stringstream cmd_ss;
		cmd_ss << hex << showbase << addr_i << " " << (byte_i & 0xff) << " wb " << flush;
		sp_tx(sp_i, cmd_ss.str());  // tx cmd
		error_f = sp_rx_wait(sp_i);  // rx wait
	}
	return(false);
}

// write preset update flag in mem
// return true if error
bool pre_upd(const sp_type& sp_i, bool usr_i) {
	int32_t data = (usr_i) ? 1 : 3;
	uint32_t addr = PRE_UPD_ADDR;
	bool error_f = mem_byte_wr(sp_i, addr, data);
	ms_sleep(PRE_UPD_MDLY);  // slow down repeated updates
	return(error_f);
}

// write knob value (to mem)
// return true if error
bool knob_wr(const sp_type& sp_i, int32_t data_i, int32_t idx_i, bool usr_i) {
	bool error_f = idx_error(idx_i, usr_i);
	if (!error_f) {
		uint32_t addr = (usr_i) ? USR_ADDR : SYS_ADDR;
		addr += (idx_i * 10) + 4;  // 5th byte, 10 byte stride
		error_f = mem_byte_wr(sp_i, addr, data_i);
	}
	return(error_f);
}

// write editor knob value (to mem)
// return true if error
bool ed_knob_wr(const sp_type& sp_i, const ed_pre_t& ed_pre_io, int32_t idx_i, bool usr_i) {
	bool error_f = idx_error(idx_i, usr_i);
	if (!error_f) {
		int32_t data = (usr_i) ? ed_pre_io.usr_enc[idx_i] : ed_pre_io.sys_enc[idx_i];
		error_f = knob_wr(sp_i, data, idx_i, usr_i);
	}
	return(error_f);
}

// editor => knobs
// return true if error
bool ed_to_knobs(const sp_type& sp_i, const ed_pre_t& ed_pre_io, bool usr_i) {
	bool error_f = false;
	uint32_t idx = 0;
	while (!error_f && (idx < PRE_PARAMS) && !idx_error(idx, usr_i)) {
		error_f = ed_knob_wr(sp_i, ed_pre_io, idx, usr_i);
		idx++;
	}
	if (!error_f) { error_f = pre_upd(sp_i, usr_i); }
	return(error_f);
}

// UI encoder read
// return true if error
bool ed_enc_rd(const ed_pre_t& ed_pre_io, int32_t &enc_o, int32_t x_i, int32_t y_i) {
	bool error_f = ed_x_error(x_i);
	error_f |= ed_y_error(y_i);
	int32_t scrn = ed_xy_to_scrn(x_i, y_i);
	int32_t param = ed_xy_to_param(x_i, y_i);
	if (!error_f) { 
		string ptr = ED_PTRS[scrn][param];
		bool usr;
		int32_t idx = ptr_to_idx(ptr, usr);
		if (idx < 0) { error_f = true; }
		else { enc_o = enc_idx_rd(ed_pre_io, idx, usr); }
	}
	return(error_f);
}

// UI encoder write
// also write knob & update
// return true if error
bool ed_enc_wr(const sp_type& sp_i, ed_pre_t& ed_pre_io, int32_t enc_i, int32_t x_i, int32_t y_i, bool upd_i) {
	bool error_f = ed_x_error(x_i);
	error_f |= ed_y_error(y_i);
	int32_t scrn = ed_xy_to_scrn(x_i, y_i);
	int32_t param = ed_xy_to_param(x_i, y_i);
	if (!error_f) { 
		string ptr = ED_PTRS[scrn][param];
		bool usr;
		int32_t idx = ptr_to_idx(ptr, usr);
		if (idx < 0) { error_f = true; }
		if (!error_f) { error_f = enc_idx_wr(ed_pre_io, enc_i, idx, usr); }
		if (!error_f) { error_f = ed_knob_wr(sp_i, ed_pre_io, idx, usr); }
		if (!error_f && upd_i) { error_f = pre_upd(sp_i, usr); }
	}
	return(error_f);
}

// UI encoder read & write delta
// return true if error
bool ed_enc_delta(const sp_type& sp_i, ed_pre_t& ed_pre_io, int32_t delta_i, int32_t x_i, int32_t y_i, bool upd_i) {
	int32_t enc;
	bool error_f = ed_enc_rd(ed_pre_io, enc, x_i, y_i);
	if (!error_f) { error_f = ed_enc_wr(sp_i, ed_pre_io, enc + delta_i, x_i, y_i, upd_i); }
	return(error_f);
}

// UI value write
// also write knob & update
// return true if error
bool ed_val_wr(const sp_type& sp_i, ed_pre_t& ed_pre_io, int32_t val_i, int32_t x_i, int32_t y_i, bool upd_i) {
	bool error_f = ed_x_error(x_i);
	error_f |= ed_y_error(y_i);
	int32_t scrn = ed_xy_to_scrn(x_i, y_i);
	int32_t param = ed_xy_to_param(x_i, y_i);
	if (!error_f) { 
		string ptr = ED_PTRS[scrn][param];
		bool usr;
		int32_t idx = ptr_to_idx(ptr, usr);
		if (idx < 0) { error_f = true; }
		if (!error_f) { 
			int32_t enc = val_i;
			uint32_t type = type_idx_rd(idx, usr);
			if (type == 0xF0 || type == 0xF1) { enc = filt_enc(val_i); } // filt freq
			else if (type == 0xF2) { enc = reson_enc(val_i); } // reson freq
			error_f = ed_enc_wr(sp_i, ed_pre_io, enc, x_i, y_i, upd_i);
		}
	}
	return(error_f);
}

// dir sv => preset sv
// return true if error
bool dir_sv_to_pre_sv(const vector<string>& dir_sv_i, vector<string>& pre_sv_o) {
	bool error_f = false;
	pre_sv_o.clear();
	uint32_t idx = 0;
	while (!error_f && (idx < dir_sv_i.size())) {
		string pre_str;
		error_f |= file_to_str((dir_sv_i[idx] + "." + DLP_EXT), pre_str);
		pre_sv_o.push_back(pre_str);
		idx++;
	}
	return(error_f);
}

// read SW version
// return true if error
bool version_rd(const sp_type& sp_i, string& str_o) {
	bool error_f = sp_open_err(sp_i);  // check port
	if (!error_f) {
		// flush input
		sp_rx_flush(sp_i);
		// read data burst to string
		string str_rx = "";
		sp_tx(sp_i, "ver ");  // tx cmd
		sp_rx_burst(sp_i, str_rx);  // rx str
		// process lines in str
		stringstream ss_rx(str_rx);
		string line = "";
		if (!getline(ss_rx, str_o)) { error_f = true; }
		if (str_o != "ver ") { error_f = true; }
		if (!getline(ss_rx, str_o)) { error_f = true; }
	}
	return(error_f);
}

// slot range s/b [-99:99]
bool slot_error_user(int32_t slot_i) {
	if ((slot_i < -99) || (slot_i > 99)) { return(true); }
	else { return(false); }
}

// slot range s/b [100:103]
bool slot_error_sys(int32_t slot_i) {
	if ((slot_i < 100) || (slot_i > 103)) { return(true); }
	else { return(false); }
}

// slot range s/b [-99:103]
bool slot_error(int32_t slot_i) {
	if ((slot_i < -99) || (slot_i > 103)) { return(true); }
	else { return(false); }
}


// process string to remove extraneous stuff
// return true if error
bool str_addr_strip(string& str_i, string& str_o) {
	stringstream ss(str_i);
	string line = "";
	int32_t lines = 0;
	str_o = "";  // clear output string
	bool error_f = false;
	bool gt_f = false;
	while (!error_f && !gt_f && getline(ss, line)) {
		bool addr_f = false;
		uint32_t idx = 0;
		while (!error_f && !gt_f && !addr_f && (idx < line.size())) {
			if (line[idx] == '>') { gt_f = true; }
			if (line[idx] == ']') { addr_f = true; }
			if (addr_f) {
				if (idx+1 < line.size()) { line = line.substr(idx+1); }
				else { error_f = true; }
			}
			idx++;
		}
		if (addr_f) {  // only lines w/ addr
			lines++;
			str_o += line;
			str_o += '\n';
		}
	}
	if (!gt_f) { error_f = true; }
	return(error_f);
}


// spi => string
// return true if error
bool spi_to_str(const sp_type& sp_i, uint32_t addr_start_i, uint32_t addr_end_i, string& str_o) {
	bool error_f = sp_open_err(sp_i);  // check port
	if (!error_f) {
		// flush input
		sp_rx_flush(sp_i);
		// read data burst to string
		string str_rx = "";
		stringstream cmd_ss;
		cmd_ss << hex << showbase << addr_start_i << " " << addr_end_i << " rs " << flush;
		sp_tx(sp_i, cmd_ss.str());  // tx cmd
		sp_rx_burst(sp_i, str_rx);  // rx str
		error_f = str_addr_strip(str_rx, str_o);  // strip cmd crap
	}
	return(error_f);
}

// slot => preset string
// return true if error
bool slot_to_pre_str(const sp_type& sp_i, int32_t slot_i, string& pre_str_o) {
	int32_t slot = slot_i;
	bool error_f = slot_error(slot);
	if (!error_f) {
		// convert slot from signed to unsigned
		if (slot < 0) { slot += 256; }
		uint32_t addr_start = slot * EEPROM_PG_BYTES;
		uint32_t addr_end = addr_start + EEPROM_PG_BYTES - 1;
		error_f = spi_to_str(sp_i, addr_start, addr_end, pre_str_o);
	}
	return(error_f);
}

// set spi csn high
// return true if error
bool spi_csn_hi(const sp_type& sp_i) {
	sp_tx(sp_i, "6 0x100 wr ");  // SPI CS
	bool error_f = sp_rx_wait(sp_i);
	return(error_f);
}

// spi write enable
// return true if error
bool spi_wr_en(const sp_type& sp_i) {
	sp_tx(sp_i, "6 6 wr ");  // SPI wr enable
	bool error_f = sp_rx_wait(sp_i);
	error_f |= spi_csn_hi(sp_i);
	return(error_f);
}

// spi write wait
// return true if error
// note: sets CSN hi
bool spi_wr_wait(const sp_type& sp_i) {
	bool error_f = spi_csn_hi(sp_i);  // SPI CS
	ms_sleep(SPI_MDLY);  // SPI wr delay
	return(error_f);
}

// spi write protect
// return true if error
// prot=true: set protected
// prot=false: set unprotected
bool spi_wr_prot(const sp_type& sp_i, bool prot_i) {
	bool error_f = spi_wr_en(sp_i);  // SPI wr enable
	sp_tx(sp_i, "6 1 wr ");  // SPI WRSR reg
	error_f |= sp_rx_wait(sp_i);
	if (prot_i) { sp_tx(sp_i, "6 0xc wr "); }  // SPI prot
	else { sp_tx(sp_i, "6 0 wr "); }  // SPI unprot
	error_f |= sp_rx_wait(sp_i);
	error_f |= spi_wr_wait(sp_i);
	return(error_f);
}

// str => spi
// return true if error
bool str_to_spi(const sp_type& sp_i, uint32_t spi_addr_i, const string& str_i) {
	bool error_f = sp_open_err(sp_i);  // check port
	if (!error_f) {
		// flush input
		sp_rx_flush(sp_i);
		// process lines in file
		stringstream ss_tx(str_i);  // so we can use getline
		string line = "";
		uint32_t lines = 0;
		uint32_t spi_addr = spi_addr_i;
		cout << flush;
		error_f |= spi_wr_prot(sp_i, false);  // unprotect
		while (!error_f && getline(ss_tx, line)) {
			stringstream cmd_ss;
			if (!(spi_addr % EEPROM_PG_BYTES)) {  // page boundary
				error_f |= spi_wr_wait(sp_i);  // SPI write wait
				error_f |= spi_wr_en(sp_i);  // SPI wr enable
				cmd_ss << hex << showbase << spi_addr << " " << flush;  // addr
			}
			if (line != "0") { cmd_ss << "0x"; }  // no 0x for zero data
			cmd_ss << line << " ws " << flush;  // data
			sp_tx(sp_i, cmd_ss.str());
			error_f |= sp_rx_wait(sp_i);
			lines++;
			spi_addr += 4;
			if (!(lines % 512)) { cout << "." << flush; }  // show activity
		}
		// final SPI CS
		error_f |= spi_wr_wait(sp_i);  // SPI write wait
		error_f |= spi_wr_prot(sp_i, true);  // protect
		if (error_f) { sp_rx_flush(sp_i); }  // flush rx if error
	}
	return(error_f);
}

// preset string => slot
// return true if error
bool pre_str_to_slot(const sp_type& sp_i, int32_t slot_i, const string& pre_str_i) {
	bool error_f = false;
	int32_t slot = slot_i;
	if (!error_f) { error_f = sp_open_err(sp_i); }  // check port
	if (!error_f) { error_f = slot_error(slot); }
	if (!error_f) {
		// convert slot from signed to unsigned
		if (slot < 0) { slot += 256; }
		if (!error_f) { error_f = str_to_spi(sp_i, slot * EEPROM_PG_BYTES, pre_str_i); }
		if (error_f) { sp_rx_flush(sp_i); }  // flush rx if error
	}
	return(error_f);
}

// preset file => slot
// return true if error
bool dlp_to_slot(const sp_type& sp_i, int32_t slot_i, const string& dlp_i) {
	bool error_f = false;
	string pre_str = "";
	if (!error_f) { error_f = file_to_str(dlp_i, pre_str); }
	if (!error_f) { error_f = pre_str_to_slot(sp_i, slot_i, pre_str); }
	return(error_f);
}

// slot => preset file
// return true if error
bool slot_to_dlp(const sp_type& sp_i, int32_t slot_i, const string& pfname_i) {
	bool error_f = false;
	string pre_str = "";
	if (!error_f) { error_f = slot_to_pre_str(sp_i, slot_i, pre_str); }
	if (!error_f) { error_f = pre_str_to_dlp(pre_str, pfname_i); }
	return(error_f);
}

// editor => slot
// return true if error
bool ed_to_slot(const sp_type& sp_i, const ed_pre_t& ed_pre_io, int32_t slot_i) {
	uint8_t pre_uba[PRE_PARAMS] = { 0 };
	string pre_str = "";
	bool usr_f = (slot_i > 99) ? false : true;  // slots > 99 are system
	ed_to_pre_uba(ed_pre_io, pre_uba, usr_f);
	pre_uba_to_pre_str(pre_uba, pre_str);
	bool error_f = pre_str_to_slot(sp_i, slot_i, pre_str);
	return(error_f);
}

// slot => editor
// return true if error
bool slot_to_ed(const sp_type& sp_i, ed_pre_t& ed_pre_io, int32_t slot_i) {
	string pre_str = "";
	uint8_t pre_uba[PRE_PARAMS] = { 0 };
	bool usr_f = (slot_i > 99) ? false : true;  // slots > 99 are system
	bool error_f = slot_to_pre_str(sp_i, slot_i, pre_str);
	if (!error_f) { error_f = pre_str_to_pre_uba(pre_uba, pre_str); }
	if (!error_f) { pre_uba_to_ed(ed_pre_io, pre_uba, usr_f); }
	return(error_f);
}

// mem dump => string
// strip off addresses, leaving 32 bit values
// return true if error
bool mem_str_rd(const sp_type& sp_i, uint32_t start_addr_i, uint32_t bytes_i, string& str_o) {
	bool error_f = false;
	if (!error_f) { error_f = sp_open_err(sp_i); }  // check port
	if (!error_f) {
		// address calcs
		uint32_t end_addr = start_addr_i + bytes_i - 1;
		// flush input
		sp_rx_flush(sp_i);
		// read data burst to string
		string str_rx = "";
		stringstream cmd_ss;
		cmd_ss << hex << showbase << start_addr_i << " " << end_addr << " rm " << flush;
		sp_tx(sp_i, cmd_ss.str());  // tx cmd
		sp_rx_burst(sp_i, str_rx);  // rx str
		error_f = str_addr_strip(str_rx, str_o);  // strip cmd crap

	}
	return(error_f);
}

// mem string => mem unsigned byte array
// return true if error
bool mem_str_to_mem_uba(const string& mem_str_i, uint8_t mem_uba_o[]) {
	bool error_f = false;
	int32_t idx = 0;
	stringstream ss(mem_str_i);
	string line = "";
	while (!error_f && getline(ss, line)) {
		uint64_t line_hex = 0;  // default
		error_f = !str_is_hex(line);
		if (!error_f) { 
			line_hex = strtoul(line.c_str(), nullptr, 16);  // STRTOUL!
			for (int32_t b=0; b<4; b++) {
				uint64_t line_byte = line_hex & 0xff;
				mem_uba_o[idx] = uint8_t(line_byte);
				line_hex >>= 8;
				idx++;
			}
		}
	}
	return(error_f);
}

// mem unsigned byte array => preset unsigned byte array
// encoder bytes are the 5th byte in the field of 10 preset bytes.
// return true if error
void mem_uba_to_pre_uba(uint8_t mem_uba_i[], uint8_t pre_uba_o[], bool usr_i) {
	uint32_t rd_idx = 0;
	uint32_t wr_idx = 0;
	uint32_t wr_bytes = 0;
	if (usr_i) { wr_bytes = USR_PARAMS; }
	else { wr_bytes = SYS_PARAMS; }
	while (wr_idx < wr_bytes) {
		if ((rd_idx % 10) == 4) {  // 5th byte
			pre_uba_o[wr_idx] = mem_uba_i[rd_idx];
			wr_idx++;
		}
		rd_idx++;
	}
	while (wr_idx < PRE_PARAMS) {  // 0 pad
		pre_uba_o[wr_idx] = 0;
		wr_idx++;
	}
	return;
}

// knobs => preset unsigned byte array
// return true if error
bool knobs_to_pre_uba(const sp_type& sp_i, uint8_t pre_uba_o[], bool usr_i) {
	bool error_f = false;
	if (!error_f) { error_f = sp_open_err(sp_i); }  // check port
	if (!error_f) {
		uint32_t bytes = (usr_i) ? USR_PARAMS : SYS_PARAMS;
		uint32_t start_addr = (usr_i) ? USR_ADDR : SYS_ADDR;
		uint8_t mem_uba[PRE_PARAMS*10] = { 0 };
		string mem_str = "";
		error_f = mem_str_rd(sp_i, start_addr, bytes * 10, mem_str);
		if (!error_f) { error_f = mem_str_to_mem_uba(mem_str, mem_uba); }
		if (!error_f) { mem_uba_to_pre_uba(mem_uba, pre_uba_o, usr_i); }
	}
	return(error_f);
}

// knobs => editor
// return true if error
bool knobs_to_ed(const sp_type& sp_i, ed_pre_t& ed_pre_io, bool usr_i) {
	uint8_t pre_uba[PRE_PARAMS] = { 0 };
	bool error_f = knobs_to_pre_uba(sp_i, pre_uba, usr_i);
	if (!error_f) { pre_uba_to_ed(ed_pre_io, pre_uba, usr_i); }
	return(error_f);
}

// slots => preset string vector
// return true if error
bool slots_to_pre_sv(const sp_type& sp_i, int32_t min_i, int32_t max_i, vector<string>& pre_sv_o) {
	bool error_f = slot_error(min_i);
	if (!error_f) { error_f = slot_error(max_i); }
	if (!error_f) { error_f = (max_i < min_i); }
	int32_t slot = min_i;
	pre_sv_o.clear();
	string pre_str;
	if (!error_f) { cout << flush; }
	while (!error_f && (slot <= max_i)) {
		error_f |= slot_to_pre_str(sp_i, slot, pre_str);
		pre_sv_o.push_back(pre_str);
		slot++;
		if (!(slot % 4)) { cout << "." << flush; }  // show activity
	}
	return(error_f);
}

// slots => preset string vector, match file names
// return true if error
bool slots_to_pre_sv_names(const sp_type& sp_i, int32_t min_i, int32_t max_i, const vector<string>& dir_sv_i, vector<string>& dl_pre_sv_o, vector<string>& dl_names_o) {
	vector<string> dir_pre_sv;
	bool error_f = dir_sv_to_pre_sv(dir_sv_i, dir_pre_sv);
	if (!error_f) { error_f = slots_to_pre_sv(sp_i, min_i, max_i, dl_pre_sv_o); }
	if (!error_f) {
		dl_names_o.clear();  // clear names
		for (uint32_t i=0; i<dl_pre_sv_o.size(); i++) {
			uint32_t idx = 0;
			if (sv_str_match(dl_pre_sv_o[i], dir_pre_sv, 0, idx)) { dl_names_o.push_back(dir_sv_i[idx]); }
			else { dl_names_o.push_back(" -?- "); }
		}
	}
	return(error_f);
}

// upload *.spi sw file
// return true if error
bool spi_pump(const sp_type& sp_i, const string& fname_i) {
	string str_tx = "";
	bool error_f = file_to_str(fname_i, str_tx);  // file => str
	if (!error_f) {
		error_f = str_to_spi(sp_i, EEPROM_SW_OFFS, str_tx);
		// reset all threads
		sp_tx(sp_i, "0 0xff000000 wr ");
		error_f |= sp_rx_wait(sp_i);
		if (error_f) { sp_rx_flush(sp_i); }  // flush rx if error
	}
	return(error_f);
}

// upload file to eeprom
// return true if error
// go to end of file
bool eeprom_pump(const sp_type& sp_i, const string& fname_i) {
	string str_tx = "";
	bool error_f = file_to_str(fname_i, str_tx);  // file => str
	if (!error_f) {
		error_f = str_to_spi(sp_i, 0, str_tx);
		// reset all threads
		sp_tx(sp_i, "0 0xff000000 wr ");
		error_f |= sp_rx_wait(sp_i);
		if (error_f) { sp_rx_flush(sp_i); }  // flush rx if error
	}
	return(error_f);
}

// download eeprom to file
// return true if error
bool eeprom_dump(const sp_type& sp_i, const string& fname_i) {
	bool error_f = false;
	bool append_f = false;
	string str_rx = "";
	int32_t addr_start = 0;
	int32_t addr_inc = 0x800;
	int32_t addr_end = EEPROM_SW_OFFS + EEPROM_SW_BYTES - 1;  // presets & sw
	while (!error_f && (addr_start < addr_end)) {
		str_rx = "";
		error_f = spi_to_str(sp_i, addr_start, addr_start+addr_inc-1, str_rx);
		if (!error_f) { error_f = str_to_file(fname_i, str_rx, append_f); }
		addr_start += addr_inc;
		append_f = true;
		cout << "." << flush;  // show activity
	}
	return(error_f);
}
